@@ -3,21 +3,21 @@
 from decimal import Decimal
 from collections import defaultdict
 
-from sql import Column
+from sql import Column, Literal, Null
 from sql.aggregate import Sum
 from sql.conditionals import Coalesce
 
 from trytond import backend
+from trytond.i18n import gettext
 from trytond.model import (ModelView, ModelSQL, DeactivableMixin, fields,
     Unique, tree)
 from trytond.wizard import Wizard, StateView, StateAction, Button
 from trytond.pyson import Eval, If, PYSONEncoder, PYSONDecoder
 from trytond.transaction import Transaction
+from trytond.tools import lstrip_wildcard
 from trytond.pool import Pool
 
-__all__ = ['Account', 'AccountDistribution',
-    'OpenChartAccountStart', 'OpenChartAccount',
-    'AnalyticAccountEntry', 'AnalyticMixin']
+from .exceptions import AccountValidationError
 
 
 class Account(
@@ -83,15 +83,6 @@ class Account(
         ('closed', 'Closed'),
         ], 'State', required=True)
     note = fields.Text('Note')
-    display_balance = fields.Selection([
-        ('debit-credit', 'Debit - Credit'),
-        ('credit-debit', 'Credit - Debit'),
-        ], 'Display Balance', required=True)
-    mandatory = fields.Boolean('Mandatory', states={
-            'invisible': Eval('type') != 'root',
-            },
-        depends=['type'],
-        help="Make this account mandatory when filling documents")
     distributions = fields.One2Many(
         'analytic_account.account.distribution', 'parent',
         "Distributions",
@@ -109,11 +100,6 @@ class Account(
         super(Account, cls).__setup__()
         cls._order.insert(0, ('code', 'ASC'))
         cls._order.insert(1, ('name', 'ASC'))
-        cls._error_messages.update({
-                'invalid_distribution': (
-                    'The distribution sum of account "%(account)s" '
-                    'is not 100%%.'),
-                })
 
     @classmethod
     def __register__(cls, module_name):
@@ -123,6 +109,9 @@ class Account(
         # Migration from 4.0: remove currency
         table.not_null_action('currency', action='remove')
 
+        # Migration from 5.0: remove display_balance
+        table.drop_column('display_balance')
+
     @staticmethod
     def default_company():
         return Transaction().context.get('company')
@@ -135,14 +124,6 @@ class Account(
     def default_state():
         return 'draft'
 
-    @staticmethod
-    def default_display_balance():
-        return 'credit-debit'
-
-    @staticmethod
-    def default_mandatory():
-        return False
-
     @classmethod
     def validate(cls, accounts):
         super(Account, cls).validate(accounts)
@@ -153,9 +134,9 @@ class Account(
         if self.type != 'distribution':
             return
         if sum((d.ratio for d in self.distributions)) != 1:
-            self.raise_user_error('invalid_distribution', {
-                    'account': self.rec_name,
-                    })
+            raise AccountValidationError(
+                gettext('analytic_account.msg_invalid_distribution',
+                    account=self.rec_name))
 
     @fields.depends('company')
     def on_change_with_currency(self, name=None):
@@ -169,9 +150,10 @@ class Account(
         return 2
 
     @fields.depends('parent', 'type',
-        '_parent_parent.root', '_parent_parent.type')
+        '_parent_parent.id', '_parent_parent.root', '_parent_parent.type')
     def on_change_parent(self):
-        if self.parent and self.type != 'root':
+        if (self.parent and self.parent.id is not None and self.parent.id > 0
+                and self.type != 'root'):
             if self.parent.type == 'root':
                 self.root = self.parent
             else:
@@ -204,13 +186,13 @@ class Account(
                 ).join(move_line, 'LEFT',
                 condition=move_line.id == line.move_line
                 ).select(table.id,
-                Sum(Coalesce(line.debit, 0) - Coalesce(line.credit, 0)),
+                Sum(Coalesce(line.credit, 0) - Coalesce(line.debit, 0)),
                 where=(table.type != 'view')
                 & table.id.in_(all_ids)
-                & (table.active == True) & line_query,
+                & (table.active == Literal(True)) & line_query,
                 group_by=table.id))
         account_sum = defaultdict(Decimal)
-        for account_id, value in cursor.fetchall():
+        for account_id, value in cursor:
             account_sum.setdefault(account_id, Decimal('0.0'))
             # SQLite uses float for SUM
             if not isinstance(value, Decimal):
@@ -225,8 +207,6 @@ class Account(
                     ])
             for child in childs:
                 balance += account_sum[child.id]
-            if account.display_balance == 'credit-debit' and balance:
-                balance *= -1
             exp = Decimal(str(10.0 ** -account.currency_digits))
             balances[account.id] = balance.quantize(exp)
         return balances
@@ -263,10 +243,10 @@ class Account(
                 ).select(*columns,
                 where=(table.type != 'view')
                 & table.id.in_(ids)
-                & (table.active == True) & line_query,
+                & (table.active == Literal(True)) & line_query,
                 group_by=table.id))
 
-        for row in cursor.fetchall():
+        for row in cursor:
             account_id = row[0]
             for i, name in enumerate(names, 1):
                 value = row[i]
@@ -289,11 +269,17 @@ class Account(
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        accounts = cls.search([('code',) + tuple(clause[1:])], limit=1)
-        if accounts:
-            return [('code',) + tuple(clause[1:])]
+        if clause[1].startswith('!') or clause[1].startswith('not '):
+            bool_op = 'AND'
         else:
-            return [(cls._rec_name,) + tuple(clause[1:])]
+            bool_op = 'OR'
+        code_value = clause[2]
+        if clause[1].endswith('like'):
+            code_value = lstrip_wildcard(clause[2])
+        return [bool_op,
+            ('code', clause[1], code_value) + tuple(clause[3:]),
+            (cls._rec_name,) + tuple(clause[1:]),
+            ]
 
     def distribute(self, amount):
         "Return a list of (account, amount) distribution"
@@ -395,16 +381,11 @@ class AnalyticAccountEntry(ModelView, ModelSQL):
         depends=['company'])
     account = fields.Many2One('analytic_account.account', 'Account',
         ondelete='RESTRICT',
-        states={
-            'required': Eval('required', False),
-            },
         domain=[
             ('root', '=', Eval('root')),
             ('type', 'in', ['normal', 'distribution']),
             ],
-        depends=['root', 'required', 'company'])
-    required = fields.Function(fields.Boolean('Required'),
-        'on_change_with_required')
+        depends=['root', 'company'])
     company = fields.Function(fields.Many2One('company.company', 'Company'),
         'on_change_with_company', searcher='search_company')
 
@@ -412,14 +393,15 @@ class AnalyticAccountEntry(ModelView, ModelSQL):
     def __register__(cls, module_name):
         pool = Pool()
         Account = pool.get('analytic_account.account')
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().connection.cursor()
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        update = transaction.connection.cursor()
 
         # Migration from 3.4: use origin as the key for One2Many
         migration_3_4 = False
         old_table = 'analytic_account_account_selection_rel'
-        if TableHandler.table_exist(old_table):
-            TableHandler.table_rename(old_table, cls._table)
+        if backend.TableHandler.table_exist(old_table):
+            backend.TableHandler.table_rename(old_table, cls._table)
             migration_3_4 = True
 
         # Don't create table before renaming
@@ -433,8 +415,8 @@ class AnalyticAccountEntry(ModelView, ModelSQL):
             cursor.execute(*account.select(account.id, account.root,
                     where=account.type != 'root'))
             entry = cls.__table__()
-            for account_id, root_id in cursor.fetchall():
-                cursor.execute(*entry.update(
+            for account_id, root_id in cursor:
+                update.execute(*entry.update(
                         columns=[entry.root],
                         values=[root_id],
                         where=entry.account == account_id))
@@ -447,7 +429,7 @@ class AnalyticAccountEntry(ModelView, ModelSQL):
         t = cls.__table__()
         cls._sql_constraints += [
             ('root_origin_uniq', Unique(t, t.origin, t.root),
-                'Only one account is allowed per analytic root and origin.'),
+                'analytic_account.msg_root_origin_unique'),
             ]
 
     @classmethod
@@ -457,17 +439,9 @@ class AnalyticAccountEntry(ModelView, ModelSQL):
     @classmethod
     def get_origin(cls):
         Model = Pool().get('ir.model')
+        get_name = Model.get_name
         models = cls._get_origin()
-        models = Model.search([
-                ('model', 'in', models),
-                ])
-        return [(None, '')] + [(m.model, m.name) for m in models]
-
-    @fields.depends('root')
-    def on_change_with_required(self, name=None):
-        if self.root:
-            return self.root.mandatory
-        return False
+        return [(None, '')] + [(m, get_name(m)) for m in models]
 
     def on_change_with_company(self, name=None):
         return None
@@ -494,7 +468,7 @@ class AnalyticAccountEntry(ModelView, ModelSQL):
 
 
 class AnalyticMixin(object):
-
+    __slots__ = ()
     analytic_accounts = fields.One2Many('analytic.account.entry', 'origin',
         'Analytic Accounts',
         size=Eval('analytic_accounts_size', 0),
@@ -502,19 +476,13 @@ class AnalyticMixin(object):
     analytic_accounts_size = fields.Function(fields.Integer(
             'Analytic Accounts Size'), 'get_analytic_accounts_size')
 
-    @classmethod
-    def __setup__(cls):
-        super(AnalyticMixin, cls).__setup__()
-        cls._error_messages.update({
-                'root_account': ('Some mandatory root account are missing '
-                    'on "%(name)s"'),
-                })
-
     @classmethod
     def __register__(cls, module_name):
         pool = Pool()
         AccountEntry = pool.get('analytic.account.entry')
-        cursor = Transaction().connection.cursor()
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        update = transaction.connection.cursor()
 
         super(AnalyticMixin, cls).__register__(module_name)
 
@@ -524,9 +492,9 @@ class AnalyticMixin(object):
             entry = AccountEntry.__table__()
             table = cls.__table__()
             cursor.execute(*table.select(table.id, table.analytic_accounts,
-                    where=table.analytic_accounts != None))
-            for line_id, selection_id in cursor.fetchall():
-                cursor.execute(*entry.update(
+                    where=table.analytic_accounts != Null))
+            for line_id, selection_id in cursor:
+                update.execute(*entry.update(
                         columns=[entry.origin],
                         values=['%s,%s' % (cls.__name__, line_id)],
                         where=entry.selection == selection_id))
@@ -551,7 +519,6 @@ class AnalyticMixin(object):
                 ])
         for account in root_accounts:
             accounts.append({
-                    'required': account.mandatory,
                     'root': account.id,
                     })
         return accounts
@@ -569,29 +536,3 @@ class AnalyticMixin(object):
     def get_analytic_accounts_size(cls, records, name):
         roots = cls.default_analytic_accounts_size()
         return {r.id: roots for r in records}
-
-    @classmethod
-    def validate(cls, analytics):
-        super(AnalyticMixin, cls).validate(analytics)
-        cls.check_roots(analytics)
-
-    @classmethod
-    def check_roots(cls, analytics):
-        "Check that all mandatory root entries are defined in entries"
-        pool = Pool()
-        Account = pool.get('analytic_account.account')
-        all_mandatory_roots = {a for a in Account.search([
-                ('type', '=', 'root'),
-                ('mandatory', '=', True),
-                ])}
-        for analytic in analytics:
-            analytic_roots = {e.root for e in analytic.analytic_accounts}
-            companies = {e.company for e in analytic.analytic_accounts}
-            mandatory_roots = set()
-            for mandatory in all_mandatory_roots:
-                if mandatory.company in companies:
-                    mandatory_roots.add(mandatory)
-            if not mandatory_roots <= analytic_roots:
-                cls.raise_user_error('root_account', {
-                        'name': analytic.rec_name,
-                        })
